"""
PyPG: A library for PPG (PhotoPlethysmoGram) processing


Documentation
-----------
Available in the docstrings for now; online soon.


Modules
-----------
Using the modules requires an explicit import. For example:
``import pypg.cycles``

    ::

    cycles
    features
    filters
    plots


Inspired By
-----------
SciPy: A scientific computing package for Python


References
----------
https://github.com/scipy/scipy

"""

import pypg.cycles
import pypg.features
import pypg.filters
import pypg.plots
